#include <stdio.h>

#define PACKED __attribute__ ((packed))
typedef struct {
	unsigned short PACKED Reserved;
	unsigned short PACKED ResType;
	unsigned short PACKED ResCount;
} NEWHEADER;

typedef struct {
	unsigned char PACKED Width;
	unsigned char PACKED Height;
	unsigned char PACKED ColorCount;
	unsigned char PACKED reserved;
} ICONRESDIR;

typedef struct {
	unsigned char PACKED Width;
	unsigned char PACKED Height;
} CURSORDIR;

typedef struct {
	unsigned short PACKED xHotSpot;
	unsigned short PACKED yHotSpot;
} LOCALHEADER;

typedef struct {
	union {
		ICONRESDIR   Icon;
		CURSORDIR    Cursor;
	};
	union {
		struct {
			unsigned short PACKED Planes;
			unsigned short PACKED BitCount;
		};
		struct {
			unsigned short PACKED xHotSpot;
			unsigned short PACKED yHotSpot;
		};
	};
	unsigned long  PACKED BytesInRes;
	unsigned long  PACKED FileOffset;
} RESDIR;

typedef struct {
	unsigned short PACKED bfType;
	unsigned long  PACKED bfSize;
	unsigned short PACKED bfReserved1;
	unsigned short PACKED bfReserved2;
	unsigned long  PACKED bfOffBits;
} BITMAPFILEHEADER;

typedef struct {
	unsigned long  PACKED biSize;
	long           PACKED biWidth;
	long           PACKED biHeight;
	unsigned short PACKED biPlanes;
	unsigned short PACKED biBitCount;
	unsigned long  PACKED biCompression;
	unsigned long  PACKED biSizeImage;
	unsigned long  PACKED biXPelsPerMeter;
	unsigned long  PACKED biYPelsPerMeter;
	unsigned long  PACKED biClrUsed;
	unsigned long  PACKED biClrImportant;
} BITMAPINFOHEADER;



int main(int argc, char **argv)
{
	NEWHEADER newheader;
	RESDIR resdir;
	LOCALHEADER localheader;
	BITMAPFILEHEADER bitmapfileheader = { ('B'|('M'<<8)), 0, 0, 0, 0 };
	BITMAPINFOHEADER bitmapinfoheader;
	FILE *fh, *fh2;
	int i, j, fpos;
	char fn[256];
	char buf[4*256];

	if (argc < 2) {
		printf("Need a filename.\n");
		return -1;
	}

	fh = fopen(argv[1], "rb");
	if (!fh) {
		printf("Error opening %s\n", argv[1]);
		return -1;
	}

	printf("Dump of %s\n", argv[1]);

	fread(&newheader, sizeof(NEWHEADER), 1, fh);
	printf("NEWHEADER:\n"
	       "\tReserved:\t%d\n"
	       "\tResType:\t%d (%s)\n"
	       "\tResCount:\t%d\n\n",
		newheader.Reserved,
		newheader.ResType, (newheader.ResType==1)?"ICON":"CURSOR",
		newheader.ResCount);

	for (i=0; i<newheader.ResCount; i++) {
		fread(&resdir, sizeof(RESDIR), 1, fh);
		if (newheader.ResType == 1) {
			printf("RESDIR:\n"
			       "\tIcon.Width:\t\t%d\n"
			       "\tIcon.Height:\t\t%d\n"
			       "\tIcon.ColorCount:\t%d\n"
			       "\tIcon.reserved:\t\t%d\n"
			       "\tPlanes:\t\t\t%d\n"
			       "\tBitCount:\t\t%d\n",
			       resdir.Icon.Width,
			       resdir.Icon.Height,
			       resdir.Icon.ColorCount,
			       resdir.Icon.reserved,
			       resdir.Planes,
			       resdir.BitCount);
		} else {
			printf("RESDIR:\n"
			       "\tCursor.Width:\t\t%d\n"
			       "\tCursor.Height:\t\t%d\n"
			       "\txHotSpot:\t\t%d\n"
			       "\tyHotSpot:\t\t%d\n",
			       resdir.Cursor.Width,
			       resdir.Cursor.Height,
			       resdir.xHotSpot,
			       resdir.yHotSpot);
		}
		printf("\tBytesInRes:\t\t%d\n"
		       "\tFileOffset:\t\t%d\n\n",
		       resdir.BytesInRes,
		       resdir.FileOffset);

		fpos = ftell(fh);

		fseek(fh, resdir.FileOffset, SEEK_SET);
		fread(&bitmapinfoheader, sizeof(BITMAPINFOHEADER), 1, fh);

		printf("\tBITMAPINFOHEADER:\n"
		       "\t\tbiSize:\t\t\t%d\n"
		       "\t\tbiWidth:\t\t%d\n"
		       "\t\tbiHeight:\t\t%d\n"
		       "\t\tbiPlanes:\t\t%d\n"
		       "\t\tbiBitCount:\t\t%d\n"
		       "\t\tbiCompression:\t\t%d\n"
		       "\t\tbiSizeImage:\t\t%d\n"
		       "\t\tbiXPelsPerMeter:\t%d\n"
		       "\t\tbiYPelsPerMeter:\t%d\n"
		       "\t\tbiClrUsed:\t\t%d\n"
		       "\t\tbiClrImportant:\t\t%d\n\n",
			bitmapinfoheader.biSize,
			bitmapinfoheader.biWidth,
			bitmapinfoheader.biHeight,
			bitmapinfoheader.biPlanes,
			bitmapinfoheader.biBitCount,
			bitmapinfoheader.biCompression,
			bitmapinfoheader.biSizeImage,
			bitmapinfoheader.biXPelsPerMeter,
			bitmapinfoheader.biYPelsPerMeter,
			bitmapinfoheader.biClrUsed,
			bitmapinfoheader.biClrImportant);

		bitmapinfoheader.biHeight >>= 1;
		sprintf(fn, "%s_%dx%dx%d_xor.bmp", argv[1], bitmapinfoheader.biWidth, bitmapinfoheader.biHeight, bitmapinfoheader.biBitCount);

		fh2 = fopen(fn, "wb");
		if (fh2) {
			bitmapfileheader.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
			bitmapfileheader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
			switch (bitmapinfoheader.biBitCount) {
				case 1:	// 2-c
					bitmapfileheader.bfSize += 4*2;
					bitmapfileheader.bfOffBits += 4*2;
					j = bitmapinfoheader.biWidth;
					if (j & 31) j = ((j >> 5) + 1) << 5;
					j = j*bitmapinfoheader.biHeight/8;
					break;

				case 4:	// 16-c
					bitmapfileheader.bfSize += 4*16;
					bitmapfileheader.bfOffBits += 4*16;
					j = bitmapinfoheader.biWidth*bitmapinfoheader.biHeight/2;
					break;

				case 8: // 256-c
					bitmapfileheader.bfSize += 4*256;
					bitmapfileheader.bfOffBits += 4*256;
					j = bitmapinfoheader.biWidth*bitmapinfoheader.biHeight;
					break;

				case 24:
					j = bitmapinfoheader.biWidth*bitmapinfoheader.biHeight*3;
					break;

				case 32:
					j = bitmapinfoheader.biWidth*bitmapinfoheader.biHeight*4;
					break;
			}
			bitmapfileheader.bfSize += j;
			bitmapinfoheader.biSizeImage = j;
			fwrite(&bitmapfileheader, sizeof(BITMAPFILEHEADER), 1, fh2);
			fwrite(&bitmapinfoheader, sizeof(BITMAPINFOHEADER), 1, fh2);

			switch (bitmapinfoheader.biBitCount) {
				case 1: // 2-c
					fwrite("\x00\x00\x00\x00", 4, 1, fh2);
					fwrite("\xff\xff\xff\x00", 4, 1, fh2);
					break;

				case 4:	// 16-c
					fread(buf, 4, 16, fh);
					fwrite(buf, 4, 16, fh2);
					break;

				case 8: // 256-c
					fread(buf, 4, 256, fh);
					fwrite(buf, 4, 256, fh2);
					break;
			}

			for (; j >= 1024; j-=1024) {
				fread(buf, 1024, 1, fh);
				fwrite(buf, 1024, 1, fh2);
			}
			if (j) {
				fread(buf, j, 1, fh);
				fwrite(buf, j, 1, fh2);
			}

			fclose(fh2);
		}

		sprintf(fn, "%s_%dx%dx%d_and.bmp", argv[1], bitmapinfoheader.biWidth, bitmapinfoheader.biHeight, bitmapinfoheader.biBitCount);

		bitmapinfoheader.biBitCount = 1;

		fh2 = fopen(fn, "wb");
		if (fh2) {
			j = bitmapinfoheader.biWidth;
			if (j & 31) j = ((j >> 5) + 1) << 5;
			j = j*bitmapinfoheader.biHeight/8;
			bitmapinfoheader.biSizeImage = j;
			bitmapfileheader.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 4*2 + j;
			bitmapfileheader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 4*2;

			fwrite(&bitmapfileheader, sizeof(BITMAPFILEHEADER), 1, fh2);
			fwrite(&bitmapinfoheader, sizeof(BITMAPINFOHEADER), 1, fh2);

			fwrite("\x00\x00\x00\x00", 4, 1, fh2);
			fwrite("\xff\xff\xff\x00", 4, 1, fh2);

			for (; j >= 1024; j-=1024) {
				fread(buf, 1024, 1, fh);
				fwrite(buf, 1024, 1, fh2);
			}
			if (j) {
				fread(buf, j, 1, fh);
				fwrite(buf, j, 1, fh2);
			}

			fclose(fh2);
		}

		fseek(fh, fpos, SEEK_SET);
	}

	fclose(fh);

	return 0;
}
